@@ -1,11 +1,11 @@
-/* $Id: raw2tiff.c,v 1.21 2006/02/07 11:08:31 dron Exp $
+/* $Id: raw2tiff.c,v 1.27 2015-06-21 01:09:10 bfriesen Exp $
  *
  * Project:  libtiff tools
  * Purpose:  Convert raw byte sequences in TIFF images
- * Author:   Andrey Kiselev, dron@remotesensing.org
+ * Author:   Andrey Kiselev, dron@ak4719.spb.edu
  *
  ******************************************************************************
- * Copyright (c) 2002, Andrey Kiselev <dron@remotesensing.org>
+ * Copyright (c) 2002, Andrey Kiselev <dron@ak4719.spb.edu>
  *
  * Permission to use, copy, modify, distribute, and sell this software and 
  * its documentation for any purpose is hereby granted without fee, provided
@@ -53,6 +53,10 @@
 # include <io.h>
 #endif
 
+#ifdef NEED_LIBPORT
+# include "libport.h"
+#endif
+
 #include "tiffio.h"
 
 #ifndef HAVE_GETOPT
@@ -101,8 +105,10 @@ main(int argc, char* argv[])
 	uint32 row, col, band;
 	int	c;
 	unsigned char *buf = NULL, *buf1 = NULL;
+#if !HAVE_DECL_OPTARG
 	extern int optind;
 	extern char* optarg;
+#endif
 
 	while ((c = getopt(argc, argv, "c:r:H:w:l:b:d:LMp:si:o:h")) != -1) {
 		switch (c) {
@@ -270,22 +276,33 @@ main(int argc, char* argv[])
 	}
 	bufsize = width * nbands * depth;
 	buf1 = (unsigned char *)_TIFFmalloc(bufsize);
-	TIFFSetField(out, TIFFTAG_ROWSPERSTRIP,
-		     TIFFDefaultStripSize(out, rowsperstrip));
+
+	rowsperstrip = TIFFDefaultStripSize(out, rowsperstrip);
+	if (rowsperstrip > length) {
+		rowsperstrip = length;
+	}
+	TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, rowsperstrip );
+
 	lseek(fd, hdr_size, SEEK_SET);		/* Skip the file header */
 	for (row = 0; row < length; row++) {
 		switch(interleaving) {
 		case BAND:			/* band interleaved data */
 			for (band = 0; band < nbands; band++) {
-				lseek(fd,
-				      hdr_size + (length*band+row)*linebytes,
-				      SEEK_SET);
+				if (lseek(fd,
+                                          hdr_size + (length*band+row)*linebytes,
+                                          SEEK_SET) == (off_t)-1) {
+                                        fprintf(stderr,
+                                                "%s: %s: scanline %lu: seek error.\n",
+                                                argv[0], argv[optind],
+                                                (unsigned long) row);
+                                        break;
+                                }
 				if (read(fd, buf, linebytes) < 0) {
 					fprintf(stderr,
-					"%s: %s: scanline %lu: Read error.\n",
-					argv[0], argv[optind],
-					(unsigned long) row);
-				break;
+                                                "%s: %s: scanline %lu: Read error.\n",
+                                                argv[0], argv[optind],
+                                                (unsigned long) row);
+                                        break;
 				}
 				if (swab)	/* Swap bytes if needed */
 					swapBytesInScanline(buf, width, dtype);
@@ -357,7 +374,10 @@ guessSize(int fd, TIFFDataType dtype, off_t hdr_size, uint32 nbands,
 	uint32	    depth = TIFFDataWidth(dtype);
 	float	    cor_coef = 0, tmp;
 
-	fstat(fd, &filestat);
+	if (fstat(fd, &filestat) == -1) {
+                fprintf(stderr, "Failed to obtain file size.\n");
+		return -1;
+        }
 
 	if (filestat.st_size < hdr_size) {
 		fprintf(stderr, "Too large header size specified.\n");
@@ -385,6 +405,7 @@ guessSize(int fd, TIFFDataType dtype, off_t hdr_size, uint32 nbands,
 
 		return 1;
 	} else if (*width == 0 && *length == 0) {
+                unsigned int fail = 0;
 		fprintf(stderr,	"Image width and height are not specified.\n");
 
 		for (w = (uint32) sqrt(imagesize / longt);
@@ -395,26 +416,46 @@ guessSize(int fd, TIFFDataType dtype, off_t hdr_size, uint32 nbands,
 				buf1 = _TIFFmalloc(scanlinesize);
 				buf2 = _TIFFmalloc(scanlinesize);
 				h = imagesize / w;
-				lseek(fd, hdr_size + (int)(h/2)*scanlinesize,
-				      SEEK_SET);
-				read(fd, buf1, scanlinesize);
-				read(fd, buf2, scanlinesize);
-				if (swab) {
-					swapBytesInScanline(buf1, w, dtype);
-					swapBytesInScanline(buf2, w, dtype);
-				}
-				tmp = (float) fabs(correlation(buf1, buf2,
-							       w, dtype));
-				if (tmp > cor_coef) {
-					cor_coef = tmp;
-					*width = w, *length = h;
-				}
+                                do {
+                                        if (lseek(fd, hdr_size + (int)(h/2)*scanlinesize,
+                                                  SEEK_SET) == (off_t)-1) {
+                                                fprintf(stderr, "seek error.\n");
+                                                fail=1;
+                                                break;
+                                        }
+                                        if (read(fd, buf1, scanlinesize) !=
+                                            (long) scanlinesize) {
+                                                fprintf(stderr, "read error.\n");
+                                                fail=1;
+                                                break;
+                                        }
+                                        if (read(fd, buf2, scanlinesize) !=
+                                            (long) scanlinesize) {
+                                                fprintf(stderr, "read error.\n");
+                                                fail=1;
+                                                break;
+                                        }
+                                        if (swab) {
+                                                swapBytesInScanline(buf1, w, dtype);
+                                                swapBytesInScanline(buf2, w, dtype);
+                                        }
+                                        tmp = (float) fabs(correlation(buf1, buf2,
+                                                                       w, dtype));
+                                        if (tmp > cor_coef) {
+                                                cor_coef = tmp;
+                                                *width = w, *length = h;
+                                        }
+                                } while (0);
 
-				_TIFFfree(buf1);
+                                _TIFFfree(buf1);
 				_TIFFfree(buf2);
 			}
 		}
 
+                if (fail) {
+                        return -1;
+                }
+
 		fprintf(stderr,
 			"Width is guessed as %lu, height is guessed as %lu.\n",
 			(unsigned long)*width, (unsigned long)*length);
@@ -633,3 +674,10 @@ usage(void)
 }
 
 /* vim: set ts=8 sts=8 sw=8 noet: */
+/*
+ * Local Variables:
+ * mode: c
+ * c-basic-offset: 8
+ * fill-column: 78
+ * End:
+ */
